// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// grav.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// declares functions for hardware aided gravity computation via GRAPE         |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_grap_h
#define included_grap_h
#ifdef  __GRAPE

namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  typedef double            GRAPE_DATA_TYPE;
  typedef GRAPE_DATA_TYPE   GRAPE_ARRAY[3];
  //===========================================================================#
  //                                                                           |
  // functions for direct summation of cell-cell and cell-self interactions    |
  //                                                                           |
  // compute the accelerations generated by N source bodies at the positions   |
  // of (up to) K sink (=test) bodies. The array of booleans indicates         |
  // whether the sink body needs the computation of its  acceleration and      |
  // potential. If the array 0, all sinks are assumed to require those.        |
  //                                                                           |
  // NOTE                                                                      |
  // These routines must ADD the accelerations and potentials computed to      |
  // those already stored in the arrays.                                       |
  //                                                                           |
  //===========================================================================#
  //                                                                           |
  // cell-cell interaction                                                     |
  //                                                                           |
  //===========================================================================#
  void mutual_grape_grav(GRAPE_DATA_TYPE ,           // I: global eps           
			 GRAPE_DATA_TYPE ,           // I: xmin                 
			 GRAPE_DATA_TYPE ,           // I: xmax                 
			 GRAPE_DATA_TYPE ,           // I: mmin                 
			 int             ,           // I: #bodies     in cell A
			 GRAPE_ARRAY    *,           // I: positions  for cell A
			 GRAPE_DATA_TYPE*,           // I: masses     for cell A
			 bool           *,           // I: sink flags for cell A
			 GRAPE_ARRAY    *,           // O: accelerations: cell A
			 GRAPE_DATA_TYPE*,           // O: potentials:    cell A
			 int             ,           // I: #bodies     in cell B
			 GRAPE_ARRAY    *,           // I: positions  for cell B
			 GRAPE_DATA_TYPE*,           // I: masses     for cell B
			 bool           *,           // I: sink flags for cell B
			 GRAPE_ARRAY    *,           // O: accelerations: cell B
			 GRAPE_DATA_TYPE*);          // O: potentials:    cell B
  //===========================================================================#
  //                                                                           |
  // cell-self interaction                                                     |
  //                                                                           |
  //===========================================================================#
  void self_grape_grav  (GRAPE_DATA_TYPE ,           // I: global eps           
			 GRAPE_DATA_TYPE ,           // I: xmin                 
			 GRAPE_DATA_TYPE ,           // I: xmax                 
			 GRAPE_DATA_TYPE ,           // I: mmin                 
			 int             ,           // I: #bodies     in cell  
			 GRAPE_ARRAY    *,           // I: positions  for cell  
			 GRAPE_DATA_TYPE*,           // I: masses     for cell  
			 bool           *,           // I: sink flags for cell  
			 GRAPE_ARRAY    *,           // O: accelerations: cell  
			 GRAPE_DATA_TYPE*);          // O: potentials:    cell  
  //////////////////////////////////////////////////////////////////////////////
}                                                    // namespace nbdy {}       
////////////////////////////////////////////////////////////////////////////////

#ifndef included_inln_h
#  include <util/inln.h>
#endif

extern "C" {
#include <gp5util.h>
}

#define  GRAPE_FUNC(F) g5_##F
namespace nbdy {
  //----------------------------------------------------------------------------
  inline
  void mutual_grape_grav(GRAPE_DATA_TYPE e,
			 GRAPE_DATA_TYPE xmin,
			 GRAPE_DATA_TYPE xmax,
			 GRAPE_DATA_TYPE mmin,
			 int             nA,
			 GRAPE_ARRAY    *xA,
			 GRAPE_DATA_TYPE*mA,
			 bool           *sA,
			 GRAPE_ARRAY    *aA,
			 GRAPE_DATA_TYPE*pA,
			 int             nB,
			 GRAPE_ARRAY    *xB,
			 GRAPE_DATA_TYPE*mB,
			 bool           *sB,
			 GRAPE_ARRAY    *aB,
			 GRAPE_DATA_TYPE*pB) 
{
  if(nA > JMEMSIZE || nB > JMEMSIZE) {
    std::cerr<<"mutual_grape_grav(): too many sources\n";
    exit(1);
  }
  // 1.  prepare                                                                
  // 1.1 figure out the maximum number of sinks                                 
  register int NsA,NsB,Nsinks;
  if(sA || sB) {
    NsA = NsB = 0;
    for(register int i=0; i!=nA; ++i) if(sA[i]) ++NsA;
    for(register int i=0; i!=nB; ++i) if(sB[i]) ++NsB;
  } else {
    NsA = nA;
    NsB = nB;
  }
  Nsinks = util::max(NsA,NsB);
  // 1.2 allocate memory for sink properties                                    
  GRAPE_ARRAY     *XS = (sA || sB)? new GRAPE_ARRAY[Nsinks] : 0;
  GRAPE_ARRAY     *AS = new GRAPE_ARRAY    [Nsinks];
  GRAPE_DATA_TYPE *PS = new GRAPE_DATA_TYPE[Nsinks];
  // 1.3 initialize GRAPE                                                       
  GRAPE_FUNC(open)();
  GRAPE_FUNC(set_range)(xmin,xmax,mmin);
  // 2   perform the interaction A -> B                                         
  if(NsB) {
    // 2.1 load source properties onto board                                    
    GRAPE_FUNC(set_mj)        (0, nA, mA);
    GRAPE_FUNC(set_xj)        (0, nA, xA);
    GRAPE_FUNC(set_eps_to_all)(e);
    GRAPE_FUNC(set_n)         (nA);
    // 2.2 load sink properties and compute gravity, add up gravity             
    if(sB) {
      for(register int j=0,i=0; i!=nB; ++i) if(sB[i]) {
	XS[j][0] = xB[i][0];
	XS[j][1] = xB[i][1];
	XS[j][2] = xB[i][2];
	++j;
      }
      GRAPE_FUNC(calculate_force_on_x)(XS,AS,PS,NsB);
      for(register int j=0,i=0; i!=nB; ++i) if(sB[i]) {
	pB[i]   += PS[j];
	aB[i][0]+= AS[j][0];
	aB[i][1]+= AS[j][1];
	aB[i][2]+= AS[j][2];
	++j;
      }
    } else {
      GRAPE_FUNC(calculate_force_on_x)(xB,AS,PS,nB);
      for(register int i=0; i!=nB; ++i) {
	pB[i]   += PS[i];
	aB[i][0]+= AS[i][0];
	aB[i][1]+= AS[i][1];
	aB[i][2]+= AS[i][2];
      }
    }
  }
  // 3   perform the interaction B -> A                                         
  if(NsB) {
    // 3.1 load source properties onto board                                    
    GRAPE_FUNC(set_mj)        (0, nB, mB);
    GRAPE_FUNC(set_xj)        (0, nB, xB);
    GRAPE_FUNC(set_eps_to_all)(e);
    GRAPE_FUNC(set_n)         (nB);
    // 3.2 load sink properties and compute gravity, add up gravity             
    if(sA) {
      for(register int j=0,i=0; i!=nA; ++i) if(sA[i]) {
	XS[j][0] = xA[i][0];
	XS[j][1] = xA[i][1];
	XS[j][2] = xA[i][2];
	++j;
      }
      GRAPE_FUNC(calculate_force_on_x)(XS,AS,PS,NsA);
      for(register int j=0,i=0; i!=nA; ++i) if(sA[i]) {
	pA[i]   += PS[j];
	aA[i][0]+= AS[j][0];
	aA[i][1]+= AS[j][1];
	aA[i][2]+= AS[j][2];
	++j;
      }
    } else {
      GRAPE_FUNC(calculate_force_on_x)(xA,AS,PS,nA);
      for(register int i=0; i!=nA; ++i) {
	pA[i]   += PS[i];
	aA[i][0]+= AS[i][0];
	aA[i][1]+= AS[i][1];
	aA[i][2]+= AS[i][2];
      }
    }
  }
  // 4   finish                                                                 
  // 4.1 release GRAPE board                                                    
  GRAPE_FUNC(close)();
  // 4.2 free occupies memory                                                   
  if(sA || sB) delete[] XS;
  delete[] AS;
  delete[] PS;
}
  //----------------------------------------------------------------------------
  inline
  void self_grape_grav(GRAPE_DATA_TYPE e,
		       GRAPE_DATA_TYPE xmin,
		       GRAPE_DATA_TYPE xmax,
		       GRAPE_DATA_TYPE mmin,
		       int             n,
		       GRAPE_ARRAY    *x,
		       GRAPE_DATA_TYPE*m,
		       bool           *s,
		       GRAPE_ARRAY    *a,
		       GRAPE_DATA_TYPE*p)
{
  if(n > JMEMSIZE) {
    std::cerr<<"self_grape_grav(): too many sources\n";
    exit(1);
  }
  // 1.  prepare                                                                
  // 1.1 figure out the maximum number of sinks                                 
  register int Nsink;
  if(s) {
    Nsink = 0;
    for(register int i=0; i!=n; ++i) if(s[i]) ++Nsink;
  } else
    Nsink = n;
  if(Nsink == 0) return;
  // 1.2 allocate memory for sink properties                                    
  GRAPE_ARRAY     *XS = s? new GRAPE_ARRAY[Nsink] : 0;
  GRAPE_ARRAY     *AS = new GRAPE_ARRAY    [Nsink];
  GRAPE_DATA_TYPE *PS = new GRAPE_DATA_TYPE[Nsink];
  // 1.3 initialize GRAPE                                                       
  GRAPE_FUNC(open)();
  GRAPE_FUNC(set_range)(xmin,xmax,mmin);
  // 2 load source properties onto board                                        
  GRAPE_FUNC(set_mj)        (0, n, m);
  GRAPE_FUNC(set_xj)        (0, n, x);
  GRAPE_FUNC(set_eps_to_all)(e);
  GRAPE_FUNC(set_n)         (n);
  // 3 load sink properties and compute gravity, add up gravity                 
  if(s) {
    for(register int j=0,i=0; i!=n; ++i) if(s[i]) {
      XS[j][0] = x[i][0];
      XS[j][1] = x[i][1];
      XS[j][2] = x[i][2];
      ++j;
    }
    GRAPE_FUNC(calculate_force_on_x)(XS,AS,PS,Nsink);
    for(register int j=0,i=0; i!=n; ++i) if(s[i]) {
      p[i]   += PS[j];
      a[i][0]+= AS[j][0];
      a[i][1]+= AS[j][1];
      a[i][2]+= AS[j][2];
      ++j;
    }
  } else {
    GRAPE_FUNC(calculate_force_on_x)(x,AS,PS,n);
    for(register int i=0; i!=n; ++i) {
      p[i]   += PS[i];
      a[i][0]+= AS[i][0];
      a[i][1]+= AS[i][1];
      a[i][2]+= AS[i][2];
    }
  }
  // 4   finish                                                                 
  // 4.1 release GRAPE board                                                    
  GRAPE_FUNC(close)();
  // 4.2 free occupies memory                                                   
  if(s) delete[] XS;
  delete[] AS;
  delete[] PS;
}
#undef GRAPE_FUNC
}      // namespace nbdy
#endif // __GRAPE
#endif // included_grap_h
