// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// tayl.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// Full support for expansion order up to p=4.                                 |
// Support for higher orders easy to implement.                                |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_tayl_h
#define included_tayl_h

#ifndef included_auxx_h
#  include <nbdy/auxx.h>
#endif

#ifndef included_soul_h
#  include <nbdy/soul.h>
#endif

#ifndef included_cell_h
#  include <nbdy/cell.h>
#endif

#ifndef included_pool_h
#  include <util/pool.h>
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
#ifdef __GRAPE
  class grape_arrays;                                // forward declaration     
#endif
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // nbdy::class TaylorExpansion                                                
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  class TaylorExpansion {
    //--------------------------------------------------------------------------
    // static methods                                                           
    //--------------------------------------------------------------------------
  public:
    static int Ncoeff(const unsigned order) {
      register int n=0,c=0,o1=order+1;
      if(NDIM==2) for(register int i=0; i!=o1; ++i) c += n += 1;
      else        for(register int i=0; i!=o1; ++i) c += n += 1+i;
      return c;
    }
    //--------------------------------------------------------------------------
    // data                                                                     
    //--------------------------------------------------------------------------
  private:
#ifdef TWODIMENSIONAL
    static const int N_C1=1, N_C2=3, N_C3=6,  N_C4=10, N_C5=15;
#else
    static const int N_C1=1, N_C2=4, N_C3=10, N_C4=20, N_C5=35;
#endif // TWODIMENSIONAL
    const   uint   PO;                              // expansion order          
    const   int    NG;                              // # coefficients per cell  
    util::pool    *POOL;                            // pool for TaylorCoeffs    
    mutable int    NC, MAXNC;                       // # TaylorCoeffs ever used 
    //--------------------------------------------------------------------------
    // struct TaylorSeries                                                      
    //--------------------------------------------------------------------------
    struct TaylorSeries {
              vect  X;                              // expansion center         
      mutable real *C;                              // coefficients             
      TaylorSeries(const vect&x, real*c) : X(x), C(c) {}
    };
    //--------------------------------------------------------------------------
    // private methods                                                          
    //--------------------------------------------------------------------------
  private:
    real& c0(real*&T) const { return *(T); }
    ten1  c1(real*&T) const { return ten1(T+N_C1); }
    ten2  c2(real*&T) const { return ten2(T+N_C2); }
    ten3  c3(real*&T) const { return ten3(T+N_C3); }
    ten4  c4(real*&T) const { return ten4(T+N_C4); }
//     ten5  c5(real*&T) const { return ten5(T+N_C5); }
    //--------------------------------------------------------------------------
    real* make_coeffs() const {
      register real*X = static_cast<real*>(POOL->alloc());
      for(register int i=0; i!=NG; ++i) X[i] = zero;
      NC++;
      return X;
    }
    //--------------------------------------------------------------------------
    real* make_coeffs(const real*Y) const { 
      register real*X = static_cast<real*>(POOL->alloc());
      for(register int i=0; i!=NG; ++i) X[i] = Y[i];
      NC++;
      return X;
    }
    //--------------------------------------------------------------------------
    void take_coeffs(real* &C) const {
      if(C) {
	POOL->free(C);
	C = 0;
	NC--;
      }
    }
    //--------------------------------------------------------------------------
    const real& c0(const TaylorSeries&T) const { return *(T.C); }
    const ten1  c1(const TaylorSeries&T) const { return ten1(T.C+N_C1); }
    const ten2  c2(const TaylorSeries&T) const { return ten2(T.C+N_C2); }
    const ten3  c3(const TaylorSeries&T) const { return ten3(T.C+N_C3); }
    const ten4  c4(const TaylorSeries&T) const { return ten4(T.C+N_C4); }
//     const ten5  c5(const TaylorSeries&T) const { return ten5(T.C+N_C5); }
    //--------------------------------------------------------------------------
    void shift_by       (real*&, const vect&) const;
    void shift_to       (TaylorSeries&, const vect&) const;
    void shift_and_add  (TaylorSeries&, const vect&,
			 const real*, const real&) const;
    void extract_grav   (const TaylorSeries&, soul*) const;
    void evaluate_grav  (cell*, const  TaylorSeries&
#ifdef __GRAPE
			 ,const grape_arrays* = 0
#endif
			 ) const;
    //--------------------------------------------------------------------------
    // protected methods                                                        
    //--------------------------------------------------------------------------
  protected:
    TaylorExpansion(const unsigned n,               // initial pool size        
		    const unsigned p=3) :           // expansion order          
      PO   ( p ),                                   // set expansion order      
      NG   ( Ncoeff(p) ),                           // size of TaylorCoeffs     
      POOL ( new util::pool(n, NG*sizeof(real)) ),  // initialize pool          
      NC   ( 0 ), MAXNC( 0 )                        // set counters             
    {
      // even though this class supports p=4, others don't. So we abort here.   
      if(PO !=3) error(" expansion orders p != 3 currently not supported");
      if(PO > 4) error(" expansion orders > 4 currently not supported");
    }
    //--------------------------------------------------------------------------
    ~TaylorExpansion() {
      delete POOL;
    }
    //--------------------------------------------------------------------------
    real& c0 (cell* C) const { return    *(C->coeffs()); }
    ten1  c1 (cell* C) const { return ten1(C->coeffs()+N_C1); }
    ten2  c2 (cell* C) const { return ten2(C->coeffs()+N_C2); }
    ten3  c3 (cell* C) const { return ten3(C->coeffs()+N_C3); }
    ten4  c4 (cell* C) const { return ten4(C->coeffs()+N_C4); }
//     ten5  c5 (cell* C) const { return ten5(C->coeffs()+N_C5); }
    //--------------------------------------------------------------------------
    void give_coeffs(cell* C) const {
      if(C->coeffs()==0) C->coeffs() = make_coeffs();
    }
    //--------------------------------------------------------------------------
    void take_coeffs(cell* C) const {
      take_coeffs(C->coeffs());
    }
    //--------------------------------------------------------------------------
#ifdef __GRAPE
    void evaluate_grav(cell*, const grape_arrays* =0) const;
#else
    void evaluate_grav(cell*) const;
#endif
    //--------------------------------------------------------------------------
    // public methods                                                           
    //--------------------------------------------------------------------------
  public:
    const  int& coeffs_used  ()      const { return MAXNC; }
    const uint& order        ()      const { return PO; }
    //--------------------------------------------------------------------------
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_tayl_h
