// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// pool.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class pool                              fully inline                        |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_pool_h
#define included_pool_h

#ifndef included_inln_h
#  include <util/inln.h>
#endif

namespace util {
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // class util::pool                                                           
  //                                                                            
  // - allocates blocks of K bytes in chunks of N elements                      
  // - single elements can be allocated and freed                               
  // - chunks are organized as linked list                                      
  // - free elements are kept in a linked list                                  
  //                                                                            
  // The actual number of bytes allocated per element is                        
  //                                                                            
  //   K' = max( K, sizeof(*void) ).                                            
  //                                                                            
  // Thus, for K < sizeof(void*) this code becomes inefficient and should not   
  // be used.                                                                   
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  class pool {
  private:
    //--------------------------------------------------------------------------
    // types of class util::pool                                                
    //--------------------------------------------------------------------------
    struct link { link *NEXT; };                     // pter to next link       
    //--------------------------------------------------------------------------
    // struct util::pool::chunk                                                 
    //--------------------------------------------------------------------------
    struct chunk {
      //........................................................................
      // data of util::pool::chunk                                              
      //........................................................................
      char   *DATA;                                  // pter to allocated memory
      chunk  *NEXT;                                  // pter to next chunk      
      //........................................................................
      // constructor                                                            
      //........................................................................
      chunk(const unsigned N,                        // I: # elements           
	    const unsigned Kp) :                     // I: size of elements     
	DATA  ( new char[N*Kp] ),                    // allocate memory         
	NEXT  ( 0 )                                  // reset pter to next chunk
      {
	const    char *END=DATA+N*Kp;                // beyond last byte        
	register char *l,*n;                         // now we link the elements
	for(l=DATA, n=DATA+Kp; n!=END; l+=Kp, n+=Kp) // loop bits of Kp bytes   
	  ((link*)(l))->NEXT = (link*)(n);           //   link to next one      
	((link*)(l))->NEXT = 0;                      // last one: null link     
      }
      //........................................................................
      // destructor                                                             
      //........................................................................
      ~chunk() { delete DATA; }                      // de-allocate memory      
    };
    //--------------------------------------------------------------------------
    // member data of class util::pool                                          
    //--------------------------------------------------------------------------
    const unsigned N;                                // # elements              
    const unsigned Kp;                               // sizeof(element)         
    chunk         *CHUNKS;                           // pter to 1st chunk       
    link          *HEAD;                             // pter to 1st free element
  
    //--------------------------------------------------------------------------
    // private methods                                                          
    //--------------------------------------------------------------------------
    void grow() {                                    // add another chunk       
      chunk *c = new chunk(N,Kp);                    //   allocate new chunk    
      c->NEXT  = CHUNKS;                             //   add to list of chunks 
      CHUNKS   = c;                                  //   make it first chunk   
      HEAD     = (link*)(CHUNKS->DATA);              //   add elements to list  
    }
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  public:
    pool(const unsigned n,                           // I: desired # elements   
	 const unsigned k) :                         // I: desired sizeof(elem) 
      N      ( max(1u,n) ),                          //  actual # elements      
      Kp     ( max(unsigned(sizeof(link)),k) ),      //  actual sizeof(element) 
      CHUNKS ( new chunk(N,Kp) ),                    //  allocate 1st chunk     
      HEAD   ( (link*)(CHUNKS->DATA) ) {}            //  get list of elements   
    //--------------------------------------------------------------------------
    // destructor                                                               
    //--------------------------------------------------------------------------
    ~pool() {
      register chunk *a=CHUNKS, *n;                  //  actual & next chunk    
      while(a) {                                     //  WHILE(actual is valid)>
	n = a->NEXT;                                 //    get next chunk       
	delete a;                                    //    delete actual chunk  
	a = n;                                       //    set actual = next    
      }                                              //  <                      
    }
    //--------------------------------------------------------------------------
    // hand out an element, i.e. a pter to Kp (=at least K) free bytes          
    //--------------------------------------------------------------------------
    void *alloc() {
      if(HEAD==0) grow();                            // no more elements: grow  
      link*p = HEAD;                                 // get 1st free element    
      HEAD   = p->NEXT;                              // reset pter to ----      
      return p;                                      // return free element     
    }
    //--------------------------------------------------------------------------
    // take back an element                                                     
    //--------------------------------------------------------------------------
    void free(void* e) {                             // I: pter to element      
      link *p = static_cast<link*>(e);               // cast into our format    
      p->NEXT = HEAD;                                // add to list of free     
      HEAD    = p;                                   // elements                
    }
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_pool_h
