// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// body.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <body.h>
#include <falcON.h>
#include <util/ionl.h>
#include <iomanip>
#include <cstring>

#ifdef ALLOW_NEMO
#include <nbdy/nmio.h>
extern "C" {
#include <stdinc.h>                                // NEMO basics               
  bool within(::real, char*, ::real);              // NEMO range checker        
}
#endif

using std::ios;
using std::cerr;
using std::setw;
using std::endl;
using std::strcmp;
using std::ostream;
using std::istream;
using namespace nbdy;
////////////////////////////////////////////////////////////////////////////////
// class bodies: NEMO I/O                                                       
////////////////////////////////////////////////////////////////////////////////
#ifdef ALLOW_NEMO
//------------------------------------------------------------------------------
io bodies::read_nemo_particles(                    // R: what has been read     
			       const nemo_in&input,// I: nemo input             
			       real         *time, //[I: read time]             
			       const io      want, //[I: what to read]          
			       char         *times,//[I: time range]            
			       const bool    warn) //[I: warn: missing data]    
{
  register uint i;                                 // index                     
  register real t;                                 // time                      
  // 1. open parameter set; read N & time; close parameter set; allocate bodies 
  input.open_set(nemo_io::param);                  // open nemo parameter set   
  register uint NN = input.read_N();               //   read N                  
  if(input.is_present(nemo_io::time))              //   IF time available      >
    input.read(nemo_io::time,&t);                  //     read time             
  else {                                           //   < ELSE >                
    t = zero;                                      //     default t=0           
    if(time && warn)                               //     IF time wanted: warn  
      nemo_io::warning("bodies::read_nemo_particles: "
		       "no time found in snapshot; defaulting to 0");
  };                                               //   <                       
  if(time) *time = t;                              //   IF wanted, set time     
  input.close_set(nemo_io::param);                 // close nemo parameter set  
  if(want == io::o ||                              // IF (nothing wanted OR     
     times &&                                      //     range given AND       
     strcmp(times,"all") != 0 &&                   //     not all times wanted  
     !within(t,times,0.0005))                      //     AND t not in range    
    return io::o;                                  //   > return undone     <   
  if(NN != N) reset(NN, want & io::e);             // IF N differs, reset       
  // 2. open bodies set                                                         
  input.open_set(nemo_io::bodies);                 // open nemo particle set    
  register io read = io::o;                        // what has been read        
  // 2.1 try to read body masses                                                
  if(want & io::m)                                 // IF(masses wanted)    >    
    if(input.is_present(nemo_io::mass)) {          //   IF(masses present, too)>
      input.read(nemo_io::mass,MASS);              //     read masses           
      read |= io::m;                               //     add mass to read      
    } else if(warn)                                //   < ELSE: warning         
      nemo_io::warning("bodies::read_nemo_particles: cannot read masses");
  // 2.2 try to read body positions and/or velocities                           
  // 2.2.1 both positions and velocities are are required                       
  if(want & io::xv) {                              // IF([x,v] wanted)     >    
    if(input.is_present(nemo_io::pos) &&           //   IF(positions and        
       input.is_present(nemo_io::vel)) {           //      velocities to read) >
      input.read(nemo_io::pos,(real*)POS);         //     read positions        
      input.read(nemo_io::vel,(real*)VEL);         //     read velocities       
      read |= io::xv;                              //     add phas to read      
    } else if(input.is_present(nemo_io::posvel)) { //   < ELSE IF(x,v present) >
      input.read(nemo_io::posvel);                 //     read phases           
      for(i=0; i!=N; ++i)                          //     loop bodies         > 
	for(register int d=0; d!=NDIM; ++d) {      //       loop dimensions   > 
	  pos(i)[d] = input.bodies_phs(i)[d];      //         copy x            
	  vel(i)[d] = input.bodies_phs(i)[d+NDIM]; //         copy v            
	}                                          //     < <                   
      read |= io::xv;                              //     add phases to read    
    } else if(warn)                                //   < ELSE: warning         
      nemo_io::warning("bodies::read_nemo_particles: cannot read phases");
  // 2.2.2 only positions are required                                          
  } else if(want & io::x) {                        // < ELSE IF( x wanted)     >
    if(input.is_present(nemo_io::pos)) {           //   IF(x present)          >
      input.read(nemo_io::pos,(real*)POS);         //     read positions        
      read |= io::x;                               //     add poss to read      
    } else if(input.is_present(nemo_io::posvel)) { //   < ELSE IF(x,v present) >
      input.read(nemo_io::posvel);                 //     read phases           
      for(i=0; i!=N; ++i)                          //     loop bodies         > 
	for(register int d=0; d!=NDIM; ++d)        //       loop dimensions   > 
	  pos(i)[d] = input.bodies_phs(i)[d];      //         copy x         < <
      read |= io::x;                               //     add poss to read      
    } else if(warn)                                //   < ELSE: warning         
      nemo_io::warning("bodies::read_nemo_particles: cannot read positions");
  // 2.2.3 only velocities are required                                         
  } else if(want & io::v) {                        // < ELSE IF( v wanted)     >
    if(input.is_present(nemo_io::vel)) {           //   IF( v present)         >
      input.read(nemo_io::vel,(real*)VEL);         //     read velocities       
      read |= io::v;                               //     add vels to read      
    } else if(input.is_present(nemo_io::posvel)) { //   < ELSE IF(x,v present) >
      input.read(nemo_io::posvel);                 //     read phases           
      for(i=0; i!=N; ++i)                          //     loop bodies         > 
	for(register int d=0; d!=NDIM; ++d)        //       loop dimensions   > 
	  vel(i)[d] = input.bodies_phs(i)[d+NDIM]; //         copy v        < < 
      read |= io::v;                               //     add vels to read      
    } else if(warn)                                //   < ELSE: warning         
      nemo_io::warning("bodies::read_nemo_particles: cannot read velocities");
  }                                                // <                         
  // 2.3 try to read accelerations                                              
  if(want & io::a)                                 // IF(acc wanted)            
    if(input.is_present(nemo_io::acc)) {           //   IF(acc present, too)    
      input.read(nemo_io::acc,(real*)(ACC));       //     read them             
      read |= io::a;                               //     add accs to read      
    } else if(warn)                                //   < ELSE: warning         
      nemo_io::warning("bodies::read_nemo_particles: cannot read accelerations");
  // 2.4 try to read potentials                                                 
  if(want & io::p)                                 // IF(pot wanted)            
    if(input.is_present(nemo_io::pot)){            //   IF(pot present, too)    
      input.read(nemo_io::pot,POT);                //     read them             
      read |= io::p;                               //     add pots to read      
    } else if(warn)                                //    < ELSE: warning        
      nemo_io::warning("bodies::read_nemo_particles: cannot read potentials");
  // 2.5 try to read densities                                                  
  if(want & io::r)                                 // IF(rho wanted)            
    if(input.is_present(nemo_io::rho)) {           //   IF(rho present, too) >  
      input.read(nemo_io::rho,RHO);                //     read them             
      read |= io::r;                               //       add rho to read     
    } else if(warn)                                //    < ELSE: warning        
      nemo_io::warning("bodies::read_nemo_particles: cannot read densities");
  // 2.6 try to read aux data                                                   
  if(want & io::y)                                 // IF(aux wanted)            
    if(input.is_present(nemo_io::aux)){            //   IF(aux present, too)    
      enable_aux_data();                           //     allocate memory       
      input.read(nemo_io::aux,AUX);                //     read them             
      read |= io::y;                               //     add auxx to read      
    } else if(warn)                                //    < ELSE: warning        
      nemo_io::warning("bodies::read_nemo_particles: cannot read aux data");
  // 2.7 try to read eps_i                                                      
  if(want & io::e)                                 // IF(eps wanted)            
    if(EPS == 0) {                                 //   IF(no memory for eps_i)>
      if(warn)                                     //     issue warning         
	nemo_io::warning("bodies::read_nemo_particles: no memory for eps_i");
    } else if(input.is_present(nemo_io::eps)){     //   ELSE IF(eps present)   >
      input.read(nemo_io::eps,EPS);                //     read them             
      read |= io::e;                               //     add epsi to read      
    } else if(warn)                                //   < ELSE: warning         
      nemo_io::warning("bodies::read_nemo_particles: cannot read eps_i");
  // finish off                                                                 
  input.close_set(nemo_io::bodies);                // close nemo particle set   
  input.reset();                                   // reset nemo input          
  return read;                                     // return add_io: what read  
}
//------------------------------------------------------------------------------
io bodies::read_nemo_snapshot (                    // R: what has been read     
			       const nemo_in&input,// I: nemo input             
			       real         *time, //[I: read time]             
			       const io      want, //[I: what to read]          
			       char         *times,//[I: time range]            
			       const bool    warn) //[I: warn: missing data]    
{
  if(!input.is_present(nemo_io::snap))             // IF no snapshot present    
    nemo_io::error("bodies::read_nemo_snapshot: no snapshot found"); // ERROR   
  input.open_set(nemo_io::snap);                   // open nemo snapshot        
  io read = read_nemo_particles(input,time,want,   // what has been read        
				times,warn);       //   by call to above?       
  input.close_set(nemo_io::snap);                  // close nemo snapshot       
  return read;                                     // return read               
}
//------------------------------------------------------------------------------
void bodies::write_nemo_particles(nemo_out  &output,
				  const real*time,
				  const io   write) const
{
  register uint i;                                 // index                     
  // 1. open parameter set; write out N & time; close parameter set;            
  output.open_set(nemo_io::param);                 // open a parameter set      
  output.write_N(N);                               //   write N                 
  if(time)                                         //   IF(time to be written)  
    output.write(nemo_io::time,*time);             //     write out time        
  output.close_set(nemo_io::param);                // close parameter set       
  // 2. allocate memory for output; open particle set; output data;             
  if(write == io::o) return;                       // nothing to be done        
  if(body_sources::N_bodies() ==0) return;         // nothing to be done        
  output.open_set(nemo_io::bodies);                // open particles set        
  output.write(nemo_io::cart);                     // put out coord system      
  // 2.1 output masses                                                          
  if(write & io::m)                                // IF(masses to be written) >
    output.write(nemo_io::mass,MASS);              //   write out masses       <
  // 2.2 - 2.3 output of positions & velocities                                 
#if(1)                  // separately as positions x_i and velocities v_i       
  // 2.2 output positions                                                       
  if(write & io::x)                                // IF poss for output   >    
    output.write(nemo_io::pos,(real*)(POS));       //   write out poss     <    
  // 2.3 output velocities                                                      
  if(write & io::v)                                // IF vels for output   >    
    output.write(nemo_io::vel,(real*)(VEL));       //   write out vels     <    
#else                   // as phase-space coordinates [x,v]_i                   
  // 2.2.1 both positions and velocities to be written out                      
  if(write & io::xv) {                             // IF([x,v] to be written)  >
    output.allocarrays();                          //   allocate memory for x,v 
    for(i=0; i!=N; ++i)                            //   loop basic bodies      >
      for(register int d=0; d!=NDIM; ++ d) {       //     loop dimensions      >
	output.bodies_phs(i)[d]      = pos(i)[d];  //       copy x              
	output.bodies_phs(i)[d+NDIM] = vel(i)[d];  //       copy v              
      }                                            //   < <                     
    output.write(nemo_io::posvel);                 //   write out phases        
  // 2.2.2 only positions to be written out                                     
  } else if(write & io::x) {                       // <ELSE IF(x to be written)>
    output.allocarrays();                          //   allocate memory for x,v 
    for(i=0; i!=N; ++i)                            //   loop basic bodies      >
      for(register int d=0; d!=NDIM; ++ d)         //     loop dimensions      >
	output.bodies_phs(i)[d]      = pos(i)[d];  //       copy x           < <
    output.write(nemo_io::posvel);                 //   write out phases        
  // 2.2.3 only velocities to be written out                                    
  } else if(write & io::v) {                       // <ELSE IF(v to be written)>
    output.allocarrays();                          //   allocate memory for x,v 
    for(i=0; i!=N; ++i)                            //   loop basic bodies      >
      for(register int d=0; d!=NDIM; ++ d)         //     loop dimensions      >
	output.bodies_phs(i)[d+NDIM] = vel(i)[d];  //       copy v           < <
    output.write(nemo_io::posvel);                 //   write out phases        
  }                                                // <                         
#endif
  // 2.4 output accelerations                                                   
  if(write & io::a)                                // IF accs for output   >    
    output.write(nemo_io::acc,(real*)(ACC));       //   write out accs     <    
  // 2.5 output potential(s)                                                    
  if(write & io::p || write & io::P) {             // IF pots for output  >     
    if(write & io::p)                              //   IF N-body pots    >     
      if(write & io::P && PEX) {                   //     IF extern pots  >     
	output.allocscalar();                      //       allocate memory     
	for(i=0; i!=N; ++i)                        //       loop bodies         
  	  output.bodies_scl(i) = pot(i) + pex(i);  //         copy potential    
	output.write(nemo_io::pot);                //       write out potential 
      } else                                       //     < ELSE >              
	output.write(nemo_io::pot,POT);            //       write out potential 
    else if(write & io::P && PEX)                  //   < ELSE IF extern pots>  
      output.write(nemo_io::pot,PEX);              //   write out potential     
  }                                                // <                         
  // 2.6 output density                                                         
  if(write & io::r)                                // IF(rhos for output)    >  
    output.write(nemo_io::rho,RHO);                //     write out scalars  <  
  // 2.7 output aux data                                                        
  if(write & io::y && AUX)                         // IF(auxx for output)    >  
    output.write(nemo_io::aux,AUX);                //   write out scalars    <  
  // 2.8 output eps_i                                                           
  if(write & io::e)                                // IF(eps_ifor output)    >  
    output.write(nemo_io::eps,EPS);                //   write out potential  <  
  // 3. close particle set & reset nemo output                                  
  output.close_set(nemo_io::bodies);               // close particles set       
  output.reset();                                  // de-allocate arrays        
}
//------------------------------------------------------------------------------
void                                               // write bodies to output    
bodies::write_nemo_snapshot (nemo_out  &output,    // I: nemo output            
			     const real*time,      //[I: write time]            
			     const io   write)     //[I: write pot/acc etc?]    
  const
{
  output.open_set(nemo_io::snap);                  // open a new nemo snapshot  
  write_nemo_particles(output,time,write);         //   write particle set      
  output.close_set(nemo_io::snap);                 // close nemo snapshot       
}
////////////////////////////////////////////////////////////////////////////////
#endif // ALLLOW_NEMO
////////////////////////////////////////////////////////////////////////////////
// class bodies: YANC I/O                                                       
////////////////////////////////////////////////////////////////////////////////
void bodies::write_yanc_ascii(ostream &out, const io write) const
{
  if(write == io::o) return;
  register uint i;
  register io   written = io::o;
  if(write & io::m)        written |= io::m;
  if(write & io::x)        written |= io::x;
  if(write & io::v)        written |= io::v;
  if(write & io::e)        written |= io::e;
  if(write & io::p)        written |= io::p;
  if(write & io::P && PEX) written |= io::P;
  if(write & io::a)        written |= io::a;
  if(write & io::r)        written |= io::r;
  if(write & io::y && AUX) written |= io::y;
  out<<written<<endl;
  for(i=0; i!=N; ++i) {
    if(written & io::m) out << mass(i) <<" ";
    if(written & io::x) out << pos (i) <<" ";
    if(written & io::v) out << vel (i) <<" ";
    if(written & io::e) out << eps (i) <<" ";
    if(written & io::p) out << pot (i) <<" ";
    if(written & io::P) out << pex (i) <<" ";
    if(written & io::a) out << acc (i) <<" ";
    if(written & io::r) out << rho (i) <<" ";
    if(written & io::y) out << aux (i) <<" ";
    out<<"\n";
  }
  out.flush();
}
//------------------------------------------------------------------------------
io bodies::read_yanc_ascii(istream& in, const io want)
{
  register io pr,read=io::o;
  { register int p; in >> p; pr=p; }
  if(pr & io::y && want & io::y) enable_aux_data();
  if(pr & io::P && want & io::P) enable_external_pot();
  if(want&io::m) { if(pr&io::m) read|=io::m; else warning("cannot read mass");}
  if(want&io::x) { if(pr&io::x) read|=io::x; else warning("cannot read pos"); }
  if(want&io::v) { if(pr&io::v) read|=io::v; else warning("cannot read vel"); }
  if(want&io::e) { if(pr&io::e) read|=io::e; else warning("cannot read eps"); }
  if(want&io::p) { if(pr&io::e) read|=io::e; else warning("cannot read pot"); }
  if(want&io::P) { if(pr&io::P) read|=io::P; else warning("cannot read pex"); }
  if(want&io::a) { if(pr&io::a) read|=io::a; else warning("cannot read acc"); }
  if(want&io::r) { if(pr&io::r) read|=io::r; else warning("cannot read rho"); }
  if(want&io::y) { if(pr&io::y) read|=io::y; else warning("cannot read aux"); }
  register real rdum;
  register vect vdum;
  for(register uint i=0; i!=N; ++i) {
    if(pr & io::m) { if(want & io::m) in>>mass(i); else in>>rdum; }
    if(pr & io::x) { if(want & io::x) in>>pos (i); else in>>vdum; }
    if(pr & io::v) { if(want & io::v) in>>vel (i); else in>>vdum; }
    if(pr & io::e) { if(want & io::e) in>>eps (i); else in>>rdum; }
    if(pr & io::e) { if(want & io::e) in>>pot (i); else in>>rdum; }
    if(pr & io::P) { if(want & io::P) in>>pex (i); else in>>rdum; }
    if(pr & io::a) { if(want & io::a) in>>acc (i); else in>>vdum; }
    if(pr & io::r) { if(want & io::r) in>>rho (i); else in>>rdum; }
    if(pr & io::y) { if(want & io::y) in>>aux (i); else in>>rdum; }
    util::SwallowRestofLine(in);
#if 0      // TEST
    if(want&io::m) cerr<<mass(i)<<"  ";
    if(want&io::x) cerr<<pos (i)<<"  ";
    if(want&io::v) cerr<<vel (i)<<"  ";
    if(want&io::e) cerr<<eps (i)<<"  ";
    if(want&io::p) cerr<<pot (i)<<"  ";
    if(want&io::P) cerr<<pex (i)<<"  ";
    if(want&io::a) cerr<<acc (i)<<"  ";
    if(want&io::r) cerr<<rho (i)<<"  ";
    if(want&io::y) cerr<<aux (i)<<"  ";
    cerr<<endl;
#endif     // TSET
  }
  return read;
}
//------------------------------------------------------------------------------
void bodies::write_yanc_binary(ostream&out, const io write) const
{
  register io written = io::o;
  if(write & io::m)        written |= io::m;
  if(write & io::x)        written |= io::x;
  if(write & io::v)        written |= io::v;
  if(write & io::e)        written |= io::e;
  if(write & io::p)        written |= io::p;
  if(write & io::P && PEX) written |= io::P;
  if(write & io::a)        written |= io::a;
  if(write & io::r)        written |= io::r;
  if(write & io::y && AUX) written |= io::y;
  int w = written;
  out.write((char *)(&w),sizeof(int));
  if(written&io::m) out.write((char *)(MASS),N*sizeof(real));
  if(written&io::x) out.write((char *)(POS) ,N*sizeof(vect));
  if(written&io::v) out.write((char *)(VEL) ,N*sizeof(vect));
  if(written&io::e) out.write((char *)(EPS) ,N*sizeof(real));
  if(written&io::p) out.write((char *)(POT) ,N*sizeof(real));
  if(written&io::P) out.write((char *)(PEX) ,N*sizeof(real));
  if(written&io::a) out.write((char *)(ACC) ,N*sizeof(vect));
  if(written&io::r) out.write((char *)(RHO) ,N*sizeof(real));
  if(written&io::y) out.write((char *)(AUX) ,N*sizeof(real));
  out.flush();
}
//------------------------------------------------------------------------------
#define READIT(IO,NAME,TYPE)					\
  if(wr&IO) 							\
    if(want&IO) in.read ((char *)(NAME),N*sizeof(TYPE));	\
    else        in.seekg(N*sizeof(TYPE),ios::cur);		\
  else if(want&IO)
void bodies::read_yanc_binary(istream&in, const io want)
{
  register io wr;
  { register int w; in.read((char*)(&w),sizeof(int)); wr=w; }
  if(wr & io::y && want & io::y) enable_aux_data();
  if(wr & io::P && want & io::P) enable_external_pot();
  if(wr & io::e  && want & io::e && EPS==0)
    error("bodies::read_yanc_binary(): cannot read eps_i: no memory allocated");
  READIT(io::m,MASS,real) error("bodies::read_yanc_binary(): cannot read mass");
  READIT(io::x, POS,vect) error("bodies::read_yanc_binary(): cannot read pos");
  READIT(io::v, VEL,vect) error("bodies::read_yanc_binary(): cannot read vel");
  READIT(io::e, EPS,real) error("bodies::read_yanc_binary(): cannot read eps");
  READIT(io::p, POT,real) error("bodies::read_yanc_binary(): cannot read pot");
  READIT(io::P, PEX,real) error("bodies::read_yanc_binary(): cannot read pex");
  READIT(io::a, ACC,vect) error("bodies::read_yanc_binary(): cannot read acc");
  READIT(io::r, RHO,real) error("bodies::read_yanc_binary(): cannot read rho");
  READIT(io::y, AUX,real) error("bodies::read_yanc_binary(): cannot read aux");
#undef READIT
}
////////////////////////////////////////////////////////////////////////////////
// class bodies: other non-constant member methods                              
////////////////////////////////////////////////////////////////////////////////
void bodies::flag_all_as_sink() {
  for(register basic_flag*Fi=FLAG; Fi!=FLAG+N; ++Fi)
    Fi->add_flag(basic_flag::SINK);
}
////////////////////////////////////////////////////////////////////////////////
